package shinh.bogoyaml;

import java.io.*;
import java.util.*;

public class BogoYaml {
    public static void dump(Object o, Writer s)
        throws IOException, YamlException {
        PrintWriter w = new PrintWriter(new BufferedWriter(s));
        w.println("--- # bogoYAML:0.1");
	dumpLocal(o, w, 0);
        w.println("");
        w.flush();
    }

    public static void dump(Object o, OutputStream s)
        throws IOException, YamlException {
	dump(o, new OutputStreamWriter(s));
    }

    public static void dump(Object o, String filename)
        throws IOException, YamlException {
	File file = new File(filename);
	dump(o, file);
    }

    public static void dump(Object o, File filename)
        throws IOException, YamlException {
        FileOutputStream s = new FileOutputStream(filename);
	dump(o, s);
        s.close();
    }

    public static Object load(Reader s) throws IOException, YamlException {
        //	debug (Internal) stream.stdout.writeLine("load done");

        BufferedReader r = new BufferedReader(s);
        String input = "";
        int c = r.read();
        while (c != -1) {
            input += (char)c;
            c = r.read();
        }

	return loadLocal(input);
    }

    public static Object load(InputStream s) throws IOException, YamlException {
        //	debug (Internal) stream.stdout.writeLine("load done");

	return load(new InputStreamReader(s));
    }

    public static Object load(String filename) throws IOException, YamlException {
        //	debug (Internal) stream.stdout.writeLine("load " ~ filename);

	File file = new File(filename);
	return load(file);
    }

    public static Object load(File filename) throws IOException, YamlException {
        //	debug (Internal) stream.stdout.writeLine("load " ~ filename);

	return load(new FileInputStream(filename));
    }

    /*
      private void p(char[] msg) {
      debug (Internal) {
      stream.stdout.writeLine(msg);
      }
      }
    */

    String input;
    int position;
    Vector stack;
    int depth;
    String key;

    BogoYaml(String i) {
        stack = new Vector();

        input = i;
        position = 0;
        stack.setSize(10);
    }

    char current() { return input.charAt(position); }
    char current(int offset) {
        return input.charAt(position+offset);
    }

    String substr(int offset) {
        return input.substring(position, position+offset);
    }
    String substr() {
        return input.substring(position);
    }

    void goForward() { position++; }
    void goBackward() { position--; }
    void goForward(int step) { position += step; }
    void goBackward(int step) { position -= step; }

    public String toString() {
        int lines = 1;
        int pos = 0;
        int linepos = 0;
        char prev = '\0';

        for (int i = 0; i < input.length(); i++) {
            char now = input.charAt(i);
            if (now == '\r' || (now == '\n' && prev != '\r')) {
                lines++;
                linepos = 0;
            }
            prev = now;
            pos++;
            linepos++;
            if (position <= pos) break;
        }

        return " in line" + lines + " pos" + (position - pos);
    }

    boolean ignoreComment() {
        if (current() == '#') {
            while (current() != '\n' && current() != '\r') {
                goForward();
            }
            if (current() == '\r' && current(1) == '\n') goForward();
            goBackward();
            return true;
        }
        return false;
    }

    boolean procLineEnd() {
        int first = position;

        while (current() == ' ' || current() == '\t' || ignoreComment()) {
            goForward();
        }

        switch (current()) {
        case '\r':
            if (current(1) == '\n') goForward();
        case '\n':
            goForward();
        case '\0':
            return true;
        default:
            position = first;
            return false;
        }
    }

    boolean isEnd() {
        return position == input.length();
    }

    void checkStackSize(int d) {
        while (d >= stack.size()) {
            stack.setSize(stack.size() * 2);
        }
    }

    void add(Object holder) {
        if (!(holder instanceof String)) {
            stack.set(depth, holder);
        }

        for (int d = depth-1; d >= 0; d--) {
            if (stack.get(d) != null) {
                if (stack.get(d) instanceof List) {
                    List a = (List)stack.get(d);
                    a.add(holder);
                    //stack[d].array = a;
                    break;
                }
                else if (stack.get(d) instanceof Map) {
                    assert(key != null);

                    Map a = (Map)stack.get(d);
                    a.put(key, holder);
                    //stack[d].assoc = a;
                    break;
                }
                else {
                    assert(false);
                }
            }
        }
    }

    void throwError(String msg) throws YamlException {
        throw new YamlException("Parser: " + msg + toString());
    }

    void throwError(String msg, Throwable cause) throws YamlException {
        throw new YamlException("Parser: " + msg + toString(), cause);
    }

    String getScalar(boolean colonEnd) throws YamlException {
        //p("get scalar");

        if (procLineEnd()) return "";

        if (current() == '\'' || current() == '"') {
            char c = current();

            goForward();

            int pos = substr().indexOf(c);

            if (pos == -1) {
                if (c == '\'') {
                    throwError("unmatched single quote");
                }
                else {
                    throwError("unmatched double quote");
                }
            }

            String ret = substr(pos);
            pos++;
            position += pos;

            if (colonEnd) {
                while (current() == ' ' || current() == '\t') {
                    goForward();
                }
            }
            else {
                if (!procLineEnd()) {
                    if (c == '\'') {
                        throwError("garbage after single quote");
                    }
                    else {
                        throwError("garbage after double quote");
                    }
                }
            }

            if (c == '"') {
                // it ends failure with "\\n", and it's slow...
                //ret = ret.replaceAll("\\a", "\a");
                ret = ret.replaceAll("\\\\b", "\b");
                ret = ret.replaceAll("\\\\f", "\f");
                ret = ret.replaceAll("\\\\n", "\n");
                ret = ret.replaceAll("\\\\r", "\r");
                ret = ret.replaceAll("\\\\t", "\t");
                //ret = ret.replaceAll("\\v", "\v");
                ret = ret.replaceAll("\\\\\\\\\"", "\"");
                ret = ret.replaceAll("\\\\ ", " ");
                ret = ret.replaceAll("\\\\\\\\", "\\");
            }

            //p("=" | ret);

            return ret;
        }
        else {
            int pos;
            if (colonEnd) {
                pos = substr().indexOf(':');
                if (pos == -1) return "";
            }
            else {
                int crpos = substr().indexOf('\r');
                int lfpos = substr().indexOf('\n');
                if (lfpos != -1 && (crpos == -1 || lfpos < crpos)) {
                    pos = lfpos;
                }
                else if (crpos != -1) {
                    pos = crpos;
                }
                else {
                    pos = input.length() - position - 1;
                }

                int sharppos = substr().indexOf('#');
                if (sharppos != -1 && sharppos < pos) {
                    pos = sharppos;
                }
            }

            String ret = substr(pos);
            position += pos;

            ret = ret.trim();
            /*
              while (ret.charAt(ret.length()-1) == ' ' ||
              ret.charAt(ret.length()-1) == '\t') {
              ret.length() = ret.length() - 1;
              }
            */

            //p("=" + ret);

            return ret;
        }
    }

    /*
      invariant {
      assert(depth < stack.size());
      }
    */

    static Object loadLocal(String input) throws YamlException {
        BogoYaml context = new BogoYaml(input);

        YamlParserState state = new YamlParserFirstLine();

        try {
            while ((state = state.parse(context)) != null) {}
        }
        catch (YamlException e) {
            throw e;
        }
        catch (Exception e) {
            context.throwError("premature end", e);
        }

        if (context.stack.size() == 0) return null;
        else return context.stack.get(0);
    }

    static void dumpLocal(Object o, PrintWriter w, int depth)
        throws YamlException {
        if (o instanceof Map) {
            Map m = (Map)o;
            for (Iterator i = m.keySet().iterator();
                 i.hasNext(); ) {
                String key = (String)i.next();
                w.println("");
                for (int j = 0; j < depth; j++) w.print('\t');
                w.print(key + ": ");
                dumpLocal(m.get(key), w, depth+1);
            }
        }
        else if (o instanceof List) {
            List l = (List)o;
            for (Iterator i = l.iterator(); i.hasNext(); ) {
                w.println("");
                for (int j = 0; j < depth; j++) w.print('\t');
                w.print("- ");
                dumpLocal(i.next(), w, depth+1);
            }
        }
        else if (o instanceof String) {
            String s = (String)o;
            //static char[] esc = "\a\b\f\n\r\t\v\"\\";
            String esc = "\b\f\n\r\t\"\\";
            for (int i = 0; i < esc.length(); i++) {
                char c = esc.charAt(i);
                if (s.indexOf(c) != -1) {
                    s = "\"" + s + "\"";
                    break;
                }
            }
            w.print(s);
        }
        else {
            throw new YamlException("cannot dumpable Object: " + o);
        }
    }

}

abstract class YamlParserState {
    public YamlParserState parse(BogoYaml c) throws YamlException {
        context = c;
        return parseLocal();
    }

    public abstract YamlParserState parseLocal() throws YamlException;

    protected void throwError(String msg) throws YamlException {
        throw new YamlException(msg + context.toString());
    }

    protected BogoYaml context;
}

class YamlParserFirstLine extends YamlParserState {
    public YamlParserState parseLocal() throws YamlException {
        while (Character.isWhitespace(context.current()) ||
               context.ignoreComment()) {
            context.goForward();
        }

        if (context.substr(3).equals("---")) {
            context.goForward(3);
        }
        else {
            throwError("garbage before document header");
        }

        if (!context.procLineEnd()) {
            throwError("garbage after document header");
        }

        if (context.isEnd()) return null;
        else return new YamlParserLineHead();
    }
}

class YamlParserLineHead extends YamlParserState {
    public YamlParserState parseLocal() throws YamlException {
        // skip emtpy line
        while (context.procLineEnd()) {
            if (context.isEnd()) return null;
        }

        int d = 0;
        while (Character.isWhitespace(context.current())) {
            context.goForward();
            d++;
        }

        context.checkStackSize(d);

        for (int i = context.depth; i > d; i--) {
            context.stack.set(i, null);
        }

        context.depth = d;

        if (context.current() == '-') return new YamlParserArray();
        else return new YamlParserAssoc();
    }
}

class YamlParserArray extends YamlParserState {
    public YamlParserState parseLocal() throws YamlException {
        if (context.stack.get(context.depth) == null) {
            context.add(new Vector());
        }

        context.goForward();

        while (context.current() == ' ' ||
               context.current() == '\t' ||
               context.ignoreComment()) {
            context.goForward();
        }

        if (!context.procLineEnd()) {
            context.depth++;
            context.add(context.getScalar(false));
            context.depth--;
        }

        return new YamlParserLineHead();
    }
}

class YamlParserAssoc extends YamlParserState {
    public YamlParserState parseLocal() throws YamlException {
        if (context.stack.get(context.depth) == null) {
            context.add(new HashMap());
        }

        context.key = context.getScalar(true);

        if (context.current() != ':') {
            throwError("colon expected");
        }

        context.goForward();

        while (context.current() == ' ' ||
               context.current() == '\t' ||
               context.ignoreComment()) {
            context.goForward();
        }

        if (!context.procLineEnd()) {
            context.depth++;
            context.add(context.getScalar(false));
            context.depth--;
        }

        return new YamlParserLineHead();
    }
}
